@@ -94,22 +94,36 @@ class CodeBlock:
                         "\ndouble __DUMMY_%(id)i;\n" % sub)  # % sub
 
 
-def failure_code(sub):
+def failure_code(sub, use_goto=True):
     """
     Code contained in sub['fail'], usually substituted for %(fail)s.
 
     It sets information about current error, then goto the code
     actually handling the failure, which is defined in struct_gen().
 
+    Parameters
+    ----------
+    sub: dict
+        Contains other code snippets that can be substituted,
+        in particular 'failure_var' and 'id'.
+    use_goto: bool, True by default
+        Include a "goto" statement to the failure label.
+        Passing False is sometimes required, in which cases we have to
+        be careful to avoid executing incorrect code.
+
     """
+    if use_goto:
+        goto_statement = 'goto __label_%(id)i;' % sub
+    else:
+        goto_statement = ''
     return '''{
-        %(failure_var)s = %(id)s;
+        %(failure_var)s = %(id)i;
         if (!PyErr_Occurred()) {
             PyErr_SetString(PyExc_RuntimeError,
                 "Unexpected error in an Op's C code. "
                 "No Python exception was set.");
-            }
-        goto __label_%(id)i;}''' % sub
+        }
+        %(goto_statement)s}''' % dict(sub, goto_statement=goto_statement)
 
 
 def failure_code_init(sub):
@@ -285,7 +299,9 @@ def struct_gen(args, struct_builders, blocks, sub):
             // now I am tired of chasing segfaults because
             // initialization code had an error and some pointer has
             // a junk value.
+            #ifndef THEANO_DONT_MEMSET_STRUCT
             memset(this, 0, sizeof(*this));
+            #endif
         }
         ~%(name)s(void) {
             cleanup();
@@ -407,7 +423,7 @@ def get_c_extract_out(r, name, sub):
     # `c_extract_out` is used to extract an output variable from
     # the compute map, to be used as pre-allocated memory for `r`
     # before its value gets computed.
-    # If the node producing `r` has `check_inputs=True`, it may
+    # If the node producing `r` has `check_input=True`, it may
     # also perform type checks on the initial value of the output,
     # so we need to pass `check_input=True` to `c_extract_out`.
     # However, that code is not used by potential clients of `r`,
@@ -751,8 +767,6 @@ class CLinker(link.Linker):
             # This ensures that, when defining functions in support code,
             # we cannot have two different functions, in different modules,
             # that have the same name.
-            # It was problematic, in particular, on Mac OS X (10.6 and 10.7)
-            # when defining CUDA kernels (with Cuda 4.2 and 5.0). See gh-1172.
             name = "node_<<<<HASH_PLACEHOLDER>>>>_%i" % node_num
             isyms = [symbol[r] for r in node.inputs]
             osyms = [symbol[r] for r in node.outputs]
@@ -898,11 +912,21 @@ class CLinker(link.Linker):
         The support code from Variables is added before the support code from Ops.This might contain duplicates.
         """
         ret = []
+        if config.cmodule.debug:
+            ret.append("""
+            #ifndef DEBUG
+            #define DEBUG
+            #endif
+            """)
         # generic support code
         for x in [y.type for y in self.variables] + [
                 y.op for y in self.node_order]:
             try:
-                ret.append(x.c_support_code())
+                support_code = x.c_support_code()
+                if isinstance(support_code, list):
+                    ret.extend(support_code)
+                else:
+                    ret.append(support_code)
             except utils.MethodNotDefined:
                 pass
         return ret
@@ -1043,7 +1067,8 @@ class CLinker(link.Linker):
                     ret += x.c_header_dirs()
             except utils.MethodNotDefined:
                 pass
-        return utils.uniq(ret)
+        # filter out empty strings/None
+        return [r for r in utils.uniq(ret) if r]
 
     def libraries(self):
         """
@@ -1085,7 +1110,8 @@ class CLinker(link.Linker):
                     ret += x.c_lib_dirs()
             except utils.MethodNotDefined:
                 pass
-        return utils.uniq(ret)
+        # filter out empty strings/None
+        return [r for r in utils.uniq(ret) if r]
 
     def __compile__(self, input_storage=None, output_storage=None,
                     storage_map=None, keep_lock=False):
@@ -1124,12 +1150,13 @@ class CLinker(link.Linker):
                 output_storage.append(map[variable])
         input_storage = tuple(input_storage)
         output_storage = tuple(output_storage)
-        thunk = self.cthunk_factory(error_storage,
-                                    input_storage,
-                                    output_storage,
-                                    storage_map,
-                                    keep_lock=keep_lock)
+        thunk, module = self.cthunk_factory(error_storage,
+                                            input_storage,
+                                            output_storage,
+                                            storage_map,
+                                            keep_lock=keep_lock)
         return (thunk,
+                module,
                 [link.Container(input, storage) for input, storage in
                  izip(self.fgraph.inputs, input_storage)],
                 [link.Container(output, storage, True) for output, storage in
@@ -1185,11 +1212,11 @@ class CLinker(link.Linker):
           first_output = ostor[0].data
         """
         init_tasks, tasks = self.get_init_tasks()
-        cthunk, in_storage, out_storage, error_storage = self.__compile__(
+        cthunk, module, in_storage, out_storage, error_storage = self.__compile__(
             input_storage, output_storage, storage_map,
             keep_lock=keep_lock)
 
-        res = _CThunk(cthunk, init_tasks, tasks, error_storage)
+        res = _CThunk(cthunk, init_tasks, tasks, error_storage, module)
         res.nodes = self.node_order
         return res, in_storage, out_storage
 
@@ -1204,17 +1231,19 @@ class CLinker(link.Linker):
         The signature has the following form:
         {{{
             'CLinker.cmodule_key', compilation args, libraries,
-            header_dirs, numpy ABI version, config md5,
+            header_dirs, numpy ABI version, config hash,
             (op0, input_signature0, output_signature0),
             (op1, input_signature1, output_signature1),
             ...
             (opK, input_signatureK, output_signatureK),
         }}}
 
+        Note that config hash now uses sha256, and not md5.
+
         The signature is a tuple, some elements of which are sub-tuples.
 
         The outer tuple has a brief header, containing the compilation options
-        passed to the compiler, the libraries to link against, an md5 hash
+        passed to the compiler, the libraries to link against, a sha256 hash
         of theano.config (for all config options where "in_c_key" is True).
         It is followed by elements for every node in the topological ordering
         of `self.fgraph`.
@@ -1273,7 +1302,7 @@ class CLinker(link.Linker):
 
     def cmodule_key_variables(self, inputs, outputs, no_recycling,
                               compile_args=None, libraries=None,
-                              header_dirs=None, insert_config_md5=True,
+                              header_dirs=None, insert_config_hash=True,
                               c_compiler=None):
 
         # Assemble a dummy fgraph using the provided inputs and outputs. It is
@@ -1296,11 +1325,11 @@ class CLinker(link.Linker):
 
         fgraph = FakeFunctionGraph(inputs, outputs)
         return self.cmodule_key_(fgraph, no_recycling, compile_args,
-                                 libraries, header_dirs, insert_config_md5,
+                                 libraries, header_dirs, insert_config_hash,
                                  c_compiler)
 
     def cmodule_key_(self, fgraph, no_recycling, compile_args=None,
-                     libraries=None, header_dirs=None, insert_config_md5=True,
+                     libraries=None, header_dirs=None, insert_config_hash=True,
                      c_compiler=None):
         """
         Do the actual computation of cmodule_key in a static method
@@ -1322,7 +1351,7 @@ class CLinker(link.Linker):
         constant_ids = dict()
         op_pos = {}  # Apply -> topological position
 
-        # First we put the header, compile_args, library names and config md5
+        # First we put the header, compile_args, library names and config hash
         # into the signature.
         sig = ['CLinker.cmodule_key']  # will be cast to tuple on return
         if compile_args is not None:
@@ -1355,8 +1384,11 @@ class CLinker(link.Linker):
         # parameters from the rest of the key. If you want to add more key
         # elements, they should be before this md5 hash if and only if they
         # can lead to a different compiled file with the same source code.
-        if insert_config_md5:
-            sig.append('md5:' + theano.configparser.get_config_md5())
+
+        # NOTE: config md5 is not using md5 hash, but sha256 instead. Function
+        # string instances of md5 will be updated at a later release.
+        if insert_config_hash:
+            sig.append('md5:' + theano.configparser.get_config_hash())
         else:
             sig.append('md5: <omitted>')
 
@@ -1414,6 +1446,8 @@ class CLinker(link.Linker):
         for node_pos, node in enumerate(order):
             if hasattr(node.op, 'c_code_cache_version_apply'):
                 version.append(node.op.c_code_cache_version_apply(node))
+            if hasattr(node.op, '__props__'):
+                version.append(node.op.__props__)
             for i in node.inputs:
                 version.append(i.type.c_code_cache_version())
             for o in node.outputs:
@@ -1601,8 +1635,7 @@ class CLinker(link.Linker):
 
         ret = module.instantiate(error_storage,
                                  *(in_storage + out_storage + orphd))
-
-        return ret
+        return ret, module
 
     def instantiate_code(self, n_args):
         code = StringIO()
@@ -1647,10 +1680,13 @@ class _CThunk(object):
         WRITEME
     error_storage
         WRITEME
+    module
+        The module that was used to compile this cthunk.
+        Mostly only useful for tests.
 
     """
 
-    def __init__(self, cthunk, init_tasks, tasks, error_storage):
+    def __init__(self, cthunk, init_tasks, tasks, error_storage, module):
         global run_cthunk
         if run_cthunk is None:
             # Lazy import to avoid compilation when importing theano.
@@ -1659,6 +1695,7 @@ class _CThunk(object):
         self.init_tasks = init_tasks
         self.tasks = tasks
         self.error_storage = error_storage
+        self.module = module
 
     def find_task(self, failure_code):
         """
@@ -1760,7 +1797,7 @@ class OpWiseCLinker(link.LocalLinker):
                  storage_map=None):
 
         # The lock will be acquired when we compile the first
-        # C code. We will keep the lock untill all the function
+        # C code. We will keep the lock until all the function
         # compilation will be finished. This allow to don't
         # require the lock when all c code are already compiled!
         orig_n_lock = getattr(get_lock, "n_lock", 0)
