@@ -1,6 +1,6 @@
 # TODO test dtype != float32
 from __future__ import absolute_import, print_function, division
-import os
+import warnings
 
 try:
     import pygpu
@@ -9,11 +9,11 @@ except ImportError:
 
 import theano
 import theano.sandbox.multinomial
-from theano import Apply, config
+from theano import Apply
 from theano.gof import Op
 
 from theano.tensor import NotScalarConstantError, get_scalar_constant_value
-from .basic_ops import as_gpuarray_variable, infer_context_name, GpuKernelBase, Kernel
+from .basic_ops import as_gpuarray_variable, infer_context_name, GpuKernelBase, Kernel, gpuarray_helper_inc_dir
 from .opt import register_opt, op_lifter, register_opt2
 from .type import GpuArrayType
 from .elemwise import GpuDimShuffle
@@ -36,14 +36,14 @@ class GPUAMultinomialFromUniform(GpuKernelBase, Op):
         return ['<numpy_compat.h>', 'gpuarray_helper.h']
 
     def c_header_dirs(self):
-        return [os.path.dirname(__file__)]
+        return [gpuarray_helper_inc_dir()]
 
     def make_node(self, pvals, unis):
-        assert unis.dtype == pvals.dtype
-        assert pvals.dtype in ['float32', 'float16', 'float64']
         ctx_name = infer_context_name(pvals, unis)
         pvals = as_gpuarray_variable(pvals, ctx_name)
         unis = as_gpuarray_variable(unis, ctx_name)
+        assert pvals.dtype in ['float32', 'float16', 'float64']
+        assert unis.dtype in ['float32', 'float16', 'float64']
 
         if pvals.ndim != 2:
             raise NotImplementedError('pvals ndim should be 2', pvals.ndim)
@@ -62,24 +62,32 @@ class GPUAMultinomialFromUniform(GpuKernelBase, Op):
 
     def gpu_kernels(self, node, name):
         out_ctype = pygpu.gpuarray.dtype_to_ctype(node.outputs[0].dtype)
-        in_ctype = pygpu.gpuarray.dtype_to_ctype(node.inputs[0].dtype)
+        pvals_ctype = pygpu.gpuarray.dtype_to_ctype(node.inputs[0].dtype)
+        unis_ctype = pygpu.gpuarray.dtype_to_ctype(node.inputs[1].dtype)
         work_ctype = pygpu.gpuarray.dtype_to_ctype(work_dtype(node.inputs[0].dtype))
         write_out_ctype = write_w(node.outputs[0].dtype)
         load_in_ctype = load_w(node.inputs[0].dtype)
-        code = """
+        code = """#include "cluda.h"
+
 KERNEL void k_multi_warp_multinomial(
     const ga_size nb_multi,
     const ga_size nb_outcomes,
-    GLOBAL_MEM %(in_ctype)s * global_pvals,
+    GLOBAL_MEM %(pvals_ctype)s *global_pvals,
+    const ga_size global_pvals_offset,
     const ga_ssize pvals_row_stride,
     const ga_ssize pvals_col_stride,
-    GLOBAL_MEM %(in_ctype)s * global_unis,
+    GLOBAL_MEM %(unis_ctype)s *global_unis,
+    const ga_size global_unis_offset,
     const ga_ssize unis_stride,
-    GLOBAL_MEM %(out_ctype)s * global_outs,
+    GLOBAL_MEM %(out_ctype)s *global_outs,
+    const ga_size global_outs_offset,
     const ga_ssize outs_row_stride,
     const ga_ssize outs_col_stride
 )
 {
+    global_pvals = (GLOBAL_MEM %(pvals_ctype)s *)(((GLOBAL_MEM char *)global_pvals) + global_pvals_offset);
+    global_unis = (GLOBAL_MEM %(unis_ctype)s *)(((GLOBAL_MEM char *)global_unis) + global_unis_offset);
+    global_outs = (GLOBAL_MEM %(out_ctype)s *)(((GLOBAL_MEM char *)global_outs) + global_outs_offset);
     // each thread takes care of one multinomial draw
     int n = LDIM_0*GID_0 + LID_0;
     if (n < nb_multi)
@@ -106,17 +114,21 @@ KERNEL void k_multi_warp_multinomial(
     }
 }
 """ % dict(out_ctype=out_ctype, write_out_ctype=write_out_ctype,
-           work_ctype=work_ctype, in_ctype=in_ctype, load_in_ctype=load_in_ctype)
+           work_ctype=work_ctype, pvals_ctype=pvals_ctype,
+           unis_ctype=unis_ctype, load_in_ctype=load_in_ctype)
         return [Kernel(
             code=code, name="k_multi_warp_multinomial",
             params=[pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.GpuArray,
+                    pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.SSIZE,
                     pygpu.gpuarray.SSIZE,
                     pygpu.gpuarray.GpuArray,
+                    pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.SSIZE,
                     pygpu.gpuarray.GpuArray,
+                    pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.SSIZE,
                     pygpu.gpuarray.SSIZE],
             flags=Kernel.get_flags(node.outputs[0].dtype),
@@ -127,10 +139,10 @@ KERNEL void k_multi_warp_multinomial(
         out, = outputs
         fail = sub['fail']
         ctx = sub['params']
-        sync = bool(config.gpuarray.sync)
         kname = self.gpu_kernels(node, name)[0].objvar
         out_typecode = pygpu.gpuarray.dtype_to_typecode(node.outputs[0].dtype)
-        in_typecode = pygpu.gpuarray.dtype_to_typecode(node.inputs[0].dtype)
+        pvals_typecode = pygpu.gpuarray.dtype_to_typecode(node.inputs[0].dtype)
+        unis_typecode = pygpu.gpuarray.dtype_to_typecode(node.inputs[1].dtype)
         s = """
         PyGpuArrayObject * pvals = %(pvals)s;
         PyGpuArrayObject * unis = %(unis)s;
@@ -192,27 +204,16 @@ KERNEL void k_multi_warp_multinomial(
 
         assert(nb_blocks*nb_threads >= nb_multi);
 
-        void *args[10];
-        ssize_t strides[5] = {
-            PyGpuArray_STRIDES(pvals)[0]/gpuarray_get_elsize(%(in_typecode)s),
-            PyGpuArray_STRIDES(pvals)[1]/gpuarray_get_elsize(%(in_typecode)s),
-            PyGpuArray_STRIDES(unis)[0]/gpuarray_get_elsize(%(in_typecode)s),
-            PyGpuArray_STRIDES(out)[0]/gpuarray_get_elsize(%(out_typecode)s),
-            PyGpuArray_STRIDES(out)[1]/gpuarray_get_elsize(%(out_typecode)s)
-        };
-        int err;
-        args[0] = (void*)&PyGpuArray_DIMS(out)[1];
-        args[1] = (void*)&PyGpuArray_DIMS(out)[0];
-        args[2] = pvals->ga.data; //PyGpuArray_DEV_DATA(pvals);
-        args[3] = (void*)&strides[0];
-        args[4] = (void*)&strides[1];
-        args[5] = unis->ga.data; //PyGpuArray_DEV_DATA(unis);
-        args[6] = (void*)&strides[2];
-        args[7] = out->ga.data; //PyGpuArray_DEV_DATA(out);
-        args[8] = (void*)&strides[3];
-        args[9] = (void*)&strides[4];
+        int err = k_multi_warp_multinomial_call(
+          1, &nb_blocks, &nb_threads, 0,
+          PyGpuArray_DIMS(out)[1], PyGpuArray_DIMS(out)[0], pvals->ga.data, pvals->ga.offset,
+          PyGpuArray_STRIDES(pvals)[0]/gpuarray_get_elsize(%(pvals_typecode)s),
+          PyGpuArray_STRIDES(pvals)[1]/gpuarray_get_elsize(%(pvals_typecode)s),
+          unis->ga.data, unis->ga.offset,
+          PyGpuArray_STRIDES(unis)[0]/gpuarray_get_elsize(%(unis_typecode)s), out->ga.data,
+          out->ga.offset, PyGpuArray_STRIDES(out)[0]/gpuarray_get_elsize(%(out_typecode)s),
+          PyGpuArray_STRIDES(out)[1]/gpuarray_get_elsize(%(out_typecode)s));
 
-        err = GpuKernel_call(&%(kname)s, 1, &nb_blocks, &nb_threads, 0, args);
         if (err != GA_NO_ERROR) {
            PyErr_Format(
                 PyExc_RuntimeError,
@@ -221,18 +222,17 @@ KERNEL void k_multi_warp_multinomial(
                 GpuKernel_error(&%(kname)s, err));
             %(fail)s;
         }
-        if(%(sync)d)
-            GpuArray_sync(&(out->ga));
+
     } // END NESTED SCOPE
         """ % locals()
 
         return s
 
     def c_code_cache_version(self):
-        return (3,)
+        return (7,)
 
 
-class GPUAMultinomialWOReplacementFromUniform(GpuKernelBase, Op):
+class GPUAChoiceFromUniform(GpuKernelBase, Op):
     """
     The output is transposed compared to MultinomialWOReplacementFromUniform.
     We must insert a Transpose op after it.
@@ -241,11 +241,17 @@ class GPUAMultinomialWOReplacementFromUniform(GpuKernelBase, Op):
 
     """
 
-    __props__ = ("odtype",)
+    __props__ = ("odtype", "replace")
 
-    def __init__(self, odtype):
+    def __init__(self, odtype, replace=False):
         Op.__init__(self)
         self.odtype = odtype
+        self.replace = replace
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
+        if "replace" not in state:
+            self.replace = False
 
     def get_params(self, node):
         return node.outputs[0].type.context
@@ -254,7 +260,7 @@ class GPUAMultinomialWOReplacementFromUniform(GpuKernelBase, Op):
         return ['<numpy_compat.h>', 'gpuarray_helper.h']
 
     def c_header_dirs(self):
-        return [os.path.dirname(__file__)]
+        return [gpuarray_helper_inc_dir()]
 
     def make_node(self, pvals, unis, n):
         assert pvals.dtype == 'float32'
@@ -281,70 +287,77 @@ class GPUAMultinomialWOReplacementFromUniform(GpuKernelBase, Op):
         return Apply(self, [pvals, unis, as_scalar(n)], [out])
 
     def gpu_kernels(self, node, name):
-        code = """
+        replace = int(self.replace)
+        code = """#include "cluda.h"
+
 KERNEL void k_multi_warp_multinomial_wor(
     const ga_size nb_multi,
     const ga_size nb_outcomes,
     const ga_size n_samples,
     GLOBAL_MEM float * global_pvals_copy,
+    const ga_size global_pvals_offset,
     const ga_ssize pvals_row_stride,
     const ga_ssize pvals_col_stride,
     GLOBAL_MEM float * global_unis,
+    const ga_size global_unis_offset,
     const ga_ssize unis_stride,
     GLOBAL_MEM ga_long * global_outs,
+    const ga_size global_outs_offset,
     const ga_ssize outs_row_stride,
     const ga_ssize outs_col_stride
 )
 {
+    global_pvals_copy = (GLOBAL_MEM float *)(((GLOBAL_MEM char *)global_pvals_copy) + global_pvals_offset);
+    global_unis = (GLOBAL_MEM float *)(((GLOBAL_MEM char *)global_unis) + global_unis_offset);
+    global_outs = (GLOBAL_MEM ga_long *)(((GLOBAL_MEM char *)global_outs) + global_outs_offset);
     // each thread takes care of one multinomial-wor n_samples-draw
     int n = LDIM_0*GID_0 + LID_0;
 
     if (n < nb_multi)
     {
+        // Sum of the remaining p_vals in global_pvals_copy[n]
+        float pvals_sum = 1.;
         for (int c = 0; c < n_samples; ++c)
         {
             float cummul = 0.;
-            bool done = false;
-            const float unis_n = global_unis[(c * nb_multi + n)*unis_stride];
+            const float unis_n = global_unis[(c * nb_multi + n)*unis_stride] * pvals_sum;
             for (ga_size m = 0; m < nb_outcomes; ++m)
             {
                 float pvals_nm = global_pvals_copy[m * pvals_col_stride + n * pvals_row_stride];
                 cummul += pvals_nm;
 
-                if (!done && unis_n < cummul)
+                if (unis_n < cummul)
                 {
-                    //write out transposed for speed.
+                    // write out transposed for speed.
                     global_outs[n * outs_col_stride +
                                 c * outs_row_stride] = m;
 
-                    global_pvals_copy[m * pvals_col_stride + n * pvals_row_stride] = 0.0;
-                    cummul -= pvals_nm;
-                    done = true;
+                    if (! %(replace)s )
+                    {
+                        global_pvals_copy[m * pvals_col_stride + n * pvals_row_stride] = 0.0;
+                        pvals_sum -= pvals_nm;
+                    }
+                    break;
                 }
             }
-            // No need to renormalize after the last samples.
-            if (c == (n_samples - 1))
-                break;
-            // parallel renormalize the multinomial
-            for (ga_int k = LID_1; k < nb_outcomes; k+=LDIM_1)
-            {
-                global_pvals_copy[k * pvals_col_stride + n * pvals_row_stride] /= cummul;
-            }
         }
     }
 }
-"""
+""" % {"replace": replace}
         return [Kernel(
             code=code, name="k_multi_warp_multinomial_wor",
             params=[pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.GpuArray,
+                    pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.SSIZE,
                     pygpu.gpuarray.SSIZE,
                     pygpu.gpuarray.GpuArray,
+                    pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.SSIZE,
                     pygpu.gpuarray.GpuArray,
+                    pygpu.gpuarray.SIZE,
                     pygpu.gpuarray.SSIZE,
                     pygpu.gpuarray.SSIZE
                     ],
@@ -354,9 +367,9 @@ KERNEL void k_multi_warp_multinomial_wor(
     def c_code(self, node, name, inp, outputs, sub):
         pvals, unis, n = inp
         out, = outputs
+        replace = int(self.replace)
         fail = sub['fail']
         ctx = sub['params']
-        sync = bool(config.gpuarray.sync)
         kname = self.gpu_kernels(node, name)[0].objvar
         s = """
     PyGpuArrayObject * pvals = %(pvals)s;
@@ -387,9 +400,12 @@ KERNEL void k_multi_warp_multinomial_wor(
         PyErr_Format(PyExc_ValueError, "unis.shape[0] != pvals.shape[0] * n");
         %(fail)s
     }
-
-    pvals_copy = pygpu_copy(pvals, GA_C_ORDER);
-
+    if (! %(replace)s) {
+        pvals_copy = pygpu_copy(pvals, GA_C_ORDER);
+    } else {
+        pvals_copy = pvals;
+        Py_INCREF(pvals_copy);
+    }
     dims[0] = n_samples;
     dims[1] = PyGpuArray_DIMS(pvals)[0];
 
@@ -430,39 +446,7 @@ KERNEL void k_multi_warp_multinomial_wor(
 
         assert(nb_blocks*nb_threads >= nb_multi);
 
-        void *args[11];
-        ssize_t strides[5] = {
-            PyGpuArray_STRIDES(pvals)[0]/sizeof(float),
-            PyGpuArray_STRIDES(pvals)[1]/sizeof(float),
-            PyGpuArray_STRIDES(unis)[0]/sizeof(float),
-            PyGpuArray_STRIDES(out)[0]/8,
-            PyGpuArray_STRIDES(out)[1]/8
-        };
-        int err;
-        args[0] = (void*)&PyGpuArray_DIMS(pvals)[0];
-        args[1] = (void*)&PyGpuArray_DIMS(pvals)[1];
-        args[2] = (void*)&n_samples;
-        args[3] = pvals_copy->ga.data; //PyGpuArray_DEV_DATA(pvals);
-        args[4] = (void*)&strides[0];
-        args[5] = (void*)&strides[1];
-        args[6] = unis->ga.data; //PyGpuArray_DEV_DATA(unis);
-        args[7] = (void*)&strides[2];
-        args[8] = out->ga.data; //PyGpuArray_DEV_DATA(out);
-        args[9] = (void*)&strides[3];
-        args[10] = (void*)&strides[4];
-
-        size_t nb_threads2[2], nb_blocks2[2];
-        nb_threads2[0] = nb_threads;
-        nb_threads2[1] = 1;
-        // If we can't schedule enough threads parallelize the renormalization.
-        // I do this because we don't always use those extra threads.
-        if (nb_threads * nb_blocks < 2048)
-            nb_threads2[1] = 1024 / nb_threads;
-
-        nb_blocks2[0] = nb_blocks;
-        nb_blocks2[1] = 1;
-
-        err = GpuKernel_call(&%(kname)s, 2, nb_blocks2, nb_threads2, 0, args);
+        int err = k_multi_warp_multinomial_wor_call(1, &nb_blocks, &nb_threads, 0, PyGpuArray_DIMS(pvals)[0], PyGpuArray_DIMS(pvals)[1], n_samples, pvals_copy->ga.data, pvals_copy->ga.offset, PyGpuArray_STRIDES(pvals)[0]/sizeof(float), PyGpuArray_STRIDES(pvals)[1]/sizeof(float), unis->ga.data, unis->ga.offset, PyGpuArray_STRIDES(unis)[0]/sizeof(float), out->ga.data, out->ga.offset, PyGpuArray_STRIDES(out)[0]/8, PyGpuArray_STRIDES(out)[1]/8);
         if (err != GA_NO_ERROR) {
            PyErr_Format(
                 PyExc_RuntimeError,
@@ -472,15 +456,14 @@ KERNEL void k_multi_warp_multinomial_wor(
            Py_DECREF(pvals_copy);
            %(fail)s;
         }
-        if(%(sync)d)
-            GpuArray_sync(&(out->ga));
+
         Py_DECREF(pvals_copy);
     } // END NESTED SCOPE
         """ % locals()
         return s
 
     def c_code_cache_version(self):
-        return (4,)
+        return (10,)
 
 
 @register_opt('fast_compile')
@@ -506,13 +489,22 @@ def local_gpua_multinomial(op, context_name, inputs, outputs):
 
 
 @register_opt('fast_compile')
-@op_lifter([theano.sandbox.multinomial.MultinomialWOReplacementFromUniform])
-@register_opt2([theano.sandbox.multinomial.MultinomialWOReplacementFromUniform], 'fast_compile')
+@op_lifter([theano.sandbox.multinomial.ChoiceFromUniform])
+@register_opt2([theano.sandbox.multinomial.ChoiceFromUniform], 'fast_compile')
 def local_gpua_multinomial_wor(op, context_name, inputs, outputs):
     # TODO : need description for function
     p, u, n = inputs
     m, = outputs
     if ((p.dtype == u.dtype == 'float32') and (m.dtype == 'int64')):
-        gpu_op = GPUAMultinomialWOReplacementFromUniform(op.odtype)
+        gpu_op = GPUAChoiceFromUniform(**op._props_dict())
         return GpuDimShuffle([False, False], [1, 0])(
             gpu_op(p, u, n))
+
+
+class GPUAMultinomialWOReplacementFromUniform(GPUAChoiceFromUniform):
+    def __init__(self, *args, **kwargs):
+        warnings.warn("GPUAMultinomialWOReplacementFromUniform is deprecated, "
+                      "use GPUAChoiceFromUniform instead.",
+                      DeprecationWarning,
+                      stacklevel=2)
+        super(GPUAMultinomialWOReplacementFromUniform, self).__init__(*args, **kwargs)
